#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""!
@file   lenscontroller.py
@date   Created on Fri Jul  7 10:02:03 2023
@author Eric van Rijswick
@company VRW Technology

This module provides the APIs for interfacing with different VA Imaging's Lens Controllers. 
"""

import serial
import struct
import time
import logging


class LensControllerAFZ:
    """!
    This class provides the LensControllerAPI, a class for interfacing with the 
    VA Imaging lens controller over a serial port. The LensControllerAPI uses the Pelco-D and 
    custom commands to control lens actions such as zooming, focusing, iris control, etc.
    """
    logger = logging.getLogger('logger')
    def __init__(self, port):
        """!
        initializes the LensControllerAPI class.
        \param[in] port: The port to use for the serial connection.
        """
        self.port = port
        self.connection = None
        self.logger.setLevel(logging.INFO)

    def connect(self):
        """!
        Opens the serial connection to the lens controller.
        """
        try:
            self.connection = serial.Serial(
                port=self.port,
                baudrate=19200,
                bytesize=serial.EIGHTBITS,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
                timeout=10
            )
            print(f"Connected successfully to {self.port}.")
        except serial.SerialException as e:
            print(f"Failed to connect to {self.port} due to: {str(e)}")

    def disconnect(self):
        """!
        Closes the serial connection to the lens controller.
        """
        if self.connection:
            self.connection.close()
            print(f"Disconnected from {self.port}.")
        else:
            print("Was not connected to any port.")

    def send_command(self, command):
        """!
        Sends a command to the lens controller and reads the response.

        @param[in] command The command to send.
        """
        if self.connection:
            self.connection.write(command)
            response = self.connection.read(1)
            if not response:
                self.logger.error(f"No response received")
                return None

            if (response[0] == 0x00):
                cmd = struct.unpack('BBBBBBB', command)
                self.logger.info(f"Command sent success: {cmd}")
            else:
                cmd = struct.unpack('BBBBBBB', command)
                print(f"Failed to send command: {cmd}")
        else:
            print("You must connect to a port first.")

    def send_custom_command_and_receive(self, command1, data1=0x01, data2=0x02, data3=0x03, data4=0x04):
        """!
        Sends a custom command to the lens controller and reads the response.

        @param[in] command1 The first command to send.
        @param[in] data1 The first piece of data to send.
        @param[in] data2 The second piece of data to send.
        @param[in] data3 The third piece of data to send.
        @param[in] data4 The fourth piece of data to send.
        """
        if self.connection:
            command = self._build_custom_command(command1, data1, data2, data3, data4)

            # Clear input buffer
            self.connection.reset_input_buffer()

            self.connection.write(command)
    
            response = self.connection.read(1)
            if not response:
               self.logger.error(f"No response received: {response}")
               return None
           
            if (response[0] == 0x00):
                cmd = struct.unpack('BBBBBBB', command)
                self.logger.info(f"Command sent success: {cmd}")
            else:
                cmd = struct.unpack('BBBBBBB', command)
                self.logger.error(f"Failed to send command: {cmd} - {response}")
                return

            response = self.connection.read(4)

            # Check if response is empty or does not contain 4 bytes
            if not response:
                self.logger.error(f"Invalid response received: {response}")
                return None
            
            if len(response) != 4:
                self.logger.error(f"length: {len(response)}")
                
            try:
                response_integer = int.from_bytes(response, 'little')
            except ValueError:
                self.logger.error(f"Failed to convert response to integer: {response}")
                return None
            
            return response_integer
        else:
            print("You must connect to a port first.")

    # ############## Pelco D Commands ############## #  
    
    def pelco_d_zoom_in(self):
        """! Send Pelco D Command - Zoom in """
        self.send_command(self._build_pelco_d_command(0x00, 0x20))

    def pelco_d_zoom_out(self):
        """! Send Pelco D Command - Zoom out """
        self.send_command(self._build_pelco_d_command(0x00, 0x40))

    def pelco_d_focus_far(self):
        """! Send Pelco D Command - Focus far """
        self.send_command(self._build_pelco_d_command(0x00, 0x80))

    def pelco_d_focus_near(self):
        """! Send Pelco D Command -  Focus near"""
        self.send_command(self._build_pelco_d_command(0x01, 0x00))

    def pelco_d_iris_open(self):
        """! Send Pelco D Command - Iris open """
        self.send_command(self._build_pelco_d_command(0x02, 0x00))

    def pelco_d_iris_close(self):
        """! Send Pelco D Command - Iris close """
        self.send_command(self._build_pelco_d_command(0x04, 0x00))

    def pelco_d_stop_all(self):
        """! Send Pelco D Command - Stop all focus, iris and zoom actions """
        self.send_command(self._build_pelco_d_command(0x00, 0x00))
        
    def pelco_d_call_preset(self, preset):
        """! 
        Send Pelco D Command - Call preset 
        @param[in] preset The preset (number [1-10]) to call
        """
        self.send_command(self._build_pelco_d_command(0x00, 0x07, data2=preset))

    def pelco_d_set_preset(self, preset):
        """! 
        Send Pelco D Command - Set preset; Set current positions under preset#
        @Note: The preset is set, but will be lost after power cycle. Call 
        custom_store_current_settings() to store the preset in the lens controller. 
        @param[in] preset The preset (number [1-10]) to call
        """
        self.send_command(self._build_pelco_d_command(0x00, 0x03, data2=preset))

    # Valid value range 0 - 2
    def pelco_d_set_zoom_speed(self, speed):
        """! 
        Send Pelco D Command - Set zoom speed; Set speed for zooming in and out
        @Note: Call custom_store_current_settings() to store the set speed in the lens controller. 
        @param[in] speed The speed value (number [0-2]) 
        """
        self.send_command(self._build_pelco_d_command(0x00, 0x25, data2=speed))

    def pelco_d_set_focus_speed(self, speed):
        """! 
        Send Pelco D Command - Set focus speed; Set speed for focussing far and near
        @Note: Call custom_store_current_settings() to store the set speed in the lens controller. 
        @param[in] speed The speed value (number [0-2]) 
        """
        self.send_command(self._build_pelco_d_command(0x00, 0x27, data2=speed))        
    # ############## Custom Commands ############## #  

    def custom_calibrate_all(self):
        """! 
        Send Custom Command - Calibrate the lens controller. This is automatically done on power on
        depending of DIP-switch 1.
        @Note: The calibration process is not te be interrupted by other commands.
        """        
        self.send_command(self._build_custom_command(0x0f))
        time.sleep(5)

    def custom_select_lens_default_afz(self):
        """! 
        Send Custom Command - Select lens type: VA Imaging AFZ Lens
        """
        self.send_command(self._build_custom_command(0x07, data4=0x01))
#        time.sleep(5)
    def custom_select_lens_AFZ2(self):
        """! 
        Send Custom Command - Select lens type: VA Imaging AFZ2 Lens
        """
        self.send_command(self._build_custom_command(0x07, data4=0x02))
#        time.sleep(5)
        
    def custom_get_zoom_position(self):
        """!
        Send Custom Command - Get zoom position
        This function retrieves the current zoom position of the lens.
        @return current zoom position as an integer.
        """
        return self.send_custom_command_and_receive(0x08)    

    def custom_set_zoom_position(self, zoom_position):
        """! 
        Send Custom Command - Set zoom position of the lens
        @param[in] zoom_position: The absolute position of the zoom lens 
        """
        if type(zoom_position) == int:
            if zoom_position < 0:
                print("negative zoom position entered.")
                return
        else:
            self.logger.error("zoom position not a number.")
            return
        
        zoom_position_bytes = zoom_position.to_bytes(4, 'big')
        self.send_command(self._build_custom_command(0x04,  zoom_position_bytes[0], zoom_position_bytes[1], zoom_position_bytes[2], zoom_position_bytes[3]))

    def custom_get_focus_position(self):
        """!
        Send Custom Command - Get focus position
        This function retrieves the current focus position of the lens.
        @return focus position as an integer.
        """
        return self.send_custom_command_and_receive(0x09)    
    
    def custom_set_focus_position(self, focus_position):
        """! 
        Send Custom Command - Set focus position of the lens
        @param[in] focus_position: The absolute position of the focus lens 
        """
        if type(focus_position) == int:
            if focus_position < 0:
                print("negative zoom position entered.")
                return
        else:
            self.logger.error("zoom position not a number.")
            return
        focus_position_bytes = focus_position.to_bytes(4, 'big')
        self.send_command(self._build_custom_command(0x05,  focus_position_bytes[0], focus_position_bytes[1], focus_position_bytes[2], focus_position_bytes[3]))
    
    def custom_get_iris_position(self):
        """!
        Send Custom Command - Get iris position
        This function retrieves the current focus position of the lens.
        @return focus position as an integer.
        """
        return int(self.send_custom_command_and_receive(0x0a)/55)
    
    def custom_set_iris_position(self, iris_position):
        """! 
        Send Custom Command - Set iris position of the lens
        @param[in] iris_position: The absolute position of the iris lens 
        """
        if type(iris_position) == int:
            iris_position = int(iris_position*55)
            if iris_position < 0:
                print("negative iris position entered.")
                self.logger.error("negative iris position entered.")
                return
        else:
            print("iris position not a number.")
            self.logger.error("iris position not a number.")
            return
    
        iris_position_bytes = iris_position.to_bytes(4, 'big')
        self.send_command(self._build_custom_command(0x06,  iris_position_bytes[0], iris_position_bytes[1], iris_position_bytes[2], iris_position_bytes[3]))

    def custom_get_lens_type(self):
        """!
        Send Custom Command - Get lens type
        This function retrieves the selected lens type
        @return lens type as an integer:
            - 0: None
            - 1: VA Imaging AFZ Lens
            - 2: VA Imaging AFZ2 Lens
        """
        return self.send_custom_command_and_receive(0x10)
                                                        
    def custom_store_current_settings(self):
        """!
        Send Custom Command - store current settings in lens controller
        This function stores all settings in the lens controller. Settings will be restored after power cycle.
        """
        self.send_command(self._build_custom_command(0x11))
                          
    # Calibration no longer uses stall detection
    def custom_disable_stall_detection(self):
        """!
        Send Custom Command - Disable stall detection for all motors
        This function disables stall detection for all motors. It serves as a solution in case motors
        have more resistance than expected. 
        """
        return self.send_custom_command_and_receive(0x13)

    def custom_get_sensitivity_mode(self):
        """!
        Send Custom Command - Get sensitivity type
        This function retrieves the sensitivity mode
        @return sensitivity mode as an integer:
            - 0: Default
            - 1: Less sensitive (for long Cable)
        """
        return self.send_custom_command_and_receive(0x0C)
    def custom_set_sensitivity_mode(self, mode):
        """! 
        Send Custom 
        Command - Set sensitivity of the lens
        @param[in] sensitivity mode as an integer:
            - 0: Default
            - 1: Less sensitive (for long Cable)
        """
        if type(mode) == int:
            if mode < 0:
                print("negative mode entered.")
                return
        else:
            self.logger.error("sensitivity mode is not a number.")
            return
        sensitivity_mode = mode.to_bytes(4, 'big')
        self.send_command(self._build_custom_command(0x0b,  sensitivity_mode[0], sensitivity_mode[1], sensitivity_mode[2], sensitivity_mode[3]))
            
    def _build_pelco_d_command(self, command1, command2, data1=0x00, data2=0x00):
        command = struct.pack('BBBBBBB', 0xFF, 0x01, command1, command2, data1, data2, (0x01 + command1 + command2 + data1 + data2) % 256)
        return command

    def _build_custom_command(self, command1, data1=0x00, data2=0x00, data3=0x00, data4=0x00):
        command = struct.pack('BBBBBBB', 0xFE, command1, data1, data2, data3, data4,  (0x01 + command1 + data1 + data2 + data3 + data4) % 256)
        return command

# Usage:
# pelco = PelcoDController("/dev/tty.usbserial-230")
# pelco.connect()
# pelco.zoom_in()
# pelco.disconnect()
    


class LensControllerFocus:
    """!
    This class provides an API for the Focus lens controller, a class for interfacing with the
    Focus lens controller over a serial port. It supports methods for moving the lens to the near and far end,
    getting and setting the current focus position and setting the speed of the focus motor.
    """
    def __init__(self, port):
        self.focus_speed = 0x00000051
        self.port = port
        self.connection = None
        
    def connect(self):
        """!
        Opens the serial connection to the lens controller.
        """
        try:
            self.connection = serial.Serial(
                port=self.port,
                baudrate=115200,
                bytesize=serial.EIGHTBITS,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
                timeout=10
            )
            print(f"Connected successfully to {self.port}.")
        except serial.SerialException as e:
            print(f"Failed to connect to {self.port} due to: {str(e)}")

    def disconnect(self):
        """!
        Closes the serial connection to the lens controller.
        """
        if self.connection:
            self.connection.close()
            print(f"Disconnected from {self.port}.")
        else:
            print("Was not connected to any port.")

    def calc_modbus_crc(self, data):
        """!
        The Focus controller uses a 16-bit modbus crc to check the integrity of the data.
        This method calculates the crc and appends it to the data.
        @param[in] data The data to send.
        """
        crc = 0xFFFF
        for byte in data:
            crc ^= byte
            for _ in range(8):
                if crc & 0x0001:
                    crc = (crc >> 1) ^ 0xA001
                else:
                    crc >>= 1
        data.append(crc & 0xFF)
        data.append((crc >> 8) & 0xFF)

    def transmit_and_receive(self, data, nof_bytes):
        """!
        Transmit data to the lens controller and receive a response.
        @param[in] data The data to send.
        @param[in] nof_bytes The number of bytes to receive.
        @return True if the transmission was successful, False otherwise.
        """
        try:
            with serial.Serial(self.port, 115200, timeout=5) as ser:
                self.calc_modbus_crc(data)
                ser.write(data)
                time.sleep(0.5)  # wait for the data to be written
                if ser.in_waiting < nof_bytes:
                    return False, None
                response = ser.read(nof_bytes)
                return True, response
        except serial.SerialException as e:
            print(f"Could not open/write to serial port: {self.port_name}. Error: {str(e)}")
            return False, None
    def transmit_data(self, data):
        """!
        Transmit data to the lens controller and check if it was received.
        @param[in] data The data to send.
        @return True if the transmission was successful, False otherwise.
        """
        try:
            with serial.Serial(self.port, 115200, timeout=5) as ser:
                self.calc_modbus_crc(data)
                ser.write(data)
                time.sleep(0.5)  # wait for the data to be written
                response = ser.read_all()
                if response == data:
                    return True
                else:
                    print(f"Received unexpected response: {response}")
                    return False
        except serial.SerialException as e:
            print(f"Could not open/write to serial port: {self.port_name}. Error: {str(e)}")
            return False

    def get_focus_position(self):
        """!
        Get the current focus position of the lens.
        This is an arbitrary value that is used to determine the current position of the lens. this value is always 0 
        directly after a power cycle. After moving the lens, the position is relative to the start position and can be negative.
        For absolute positioning move the lens to the near and far end and store the positions by calling [get_end_positions](#api.LensControllerFocus.get_end_positions).
        @return The current focus position as an integer, -1 if there was an error.
        """
        data = bytearray([0x01, 0x65, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00])
        response = bytearray(16)
        result, response = self.transmit_and_receive(data, 16)
        if result:
            sub_array = response[8:12]
            rotating = response[3]
            if len(sub_array) >= 4:
                position = int.from_bytes(sub_array[:4], byteorder='big', signed=True)
  #              position = (sub_array[0] << 24) | (sub_array[1] << 16) | (sub_array[2] << 8) | sub_array[3]
                return position, rotating
            else:
                print("Error: bytearray does not contain enough bytes.")
        return -1, None

    def focus_stop(self):
        """!
        Stop the focus motor.
        """
        data = bytearray([0x01, 0x64, 0x01, 0x00, 0x00, 0x00, 0x51, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00])
        self.transmit_data(data)

    def set_focus_speed(self, rpm):
        """!
        Set the speed of the focus motor.
        @param[in] rpm The speed in revolutions per minute.
        """
        self.focus_speed = (rpm * 16384) // 6000

    def get_focus_speed(self):
        """!
        Get the speed of the focus motor.
        @return The speed in revolutions per minute.
        """
        return (self.focus_speed * 6000) // 16384
    
    def set_focus_position(self, position):
        """!
        Set the focus position of the lens.
        @param[in] position The absolute position of the focus lens.
        """
        currentPosition, rotating = self.get_focus_position()
        if rotating != 0x00:
            self.focus_stop()
            time.sleep(0.5)  # sleep for 500 milliseconds
            currentPosition, rotating = self.get_focus_position()

        speed_value = self.focus_speed
        angle_value = position - currentPosition

        data = bytearray([0x01, 0x64, 0x01])
        data.extend((speed_value >> i & 0xff for i in (24, 16, 8, 0)))  # speed
        data.extend((angle_value >> i & 0xff for i in (24, 16, 8, 0)))  # angle
        data.extend([0x00, 0x00, 0x00])

        self.transmit_data(data)

    def move_focus_position(self, steps):
        """!
        Move the focus position of the lens by an amount of steps.
        @param[in] steps The number of steps to move the lens.
        """
        speed_value = self.focus_speed
        angle_value = steps

        data = bytearray([0x01, 0x64, 0x01])
        data.extend((speed_value >> i & 0xff for i in (24, 16, 8, 0)))  # speed
        data.extend((angle_value >> i & 0xff for i in (24, 16, 8, 0)))  # angle
        data.extend([0x00, 0x00, 0x00])

        self.transmit_data(data)

    def focus_near(self):
        """!
        Move the focus position of the lens to the near end.
        """

        # Now check every x period to see if the motor is still moving
        moving = True
        position, rotating = self.get_focus_position()
        previousPosition = position
        bandwidth = 50

        while moving:
            self.move_focus_position(-48500)
            time.sleep(0.5)
            position, rotating = self.get_focus_position()
            if (position < (previousPosition + bandwidth)) and (position > (previousPosition - bandwidth)):
                moving = False
            else:
                previousPosition = position

        # Make sure the motor is off as the hw controller sometimes stays running
        self.focus_stop()

    def focus_far(self):
        """!
        Move the focus position of the lens to the far end.
        """
        # Now check every x period to see if the motor is still moving
        moving = True
        position, rotating = self.get_focus_position()
        previousPosition = position
        bandwidth = 50

        while moving:
            self.move_focus_position(48500)
            time.sleep(0.5)
            position, rotating = self.get_focus_position()
            if (position < (previousPosition + bandwidth)) and (position > (previousPosition - bandwidth)):
                moving = False
            else:
                previousPosition = position

        # Make sure the motor is off as the hw controller sometimes stays running
        self.focus_stop()

    def get_end_positions(self):
        """!
        Get the near and far end positions of the lens.
        @return The near and far end positions as a tuple.
        """
        self.focus_near()
        nearPosition, rotating = self.get_focus_position()
        self.focus_far()
        farPosition, rotating = self.get_focus_position()
        return nearPosition, farPosition
